# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

import logging
from typing import List
from django.db.models import Q
from django.utils.translation import ugettext as _
from EPCPyYes.core.v1_2 import template_events, events as pyyes_events
from EPCPyYes.core.v1_2.CBV.instance_lot_master_data import \
    InstanceLotMasterDataAttribute
from EPCPyYes.core.SBDH import sbdh, template_sbdh
from quartet_epcis.models.choices import EventTypeChoicesEnum
from quartet_epcis.models import events, entries, headers
from quartet_epcis.parsing import errors

logger = logging.getLogger(__name__)

EntryList = List[entries.Entry]


def get_sources(db_event: events.Event):
    '''
    Returns each of the source events associated with the db_event
    parameter.

    :param: db_event: The event used to look up sources.
    :return: a QuerySet containing the source model instances associated
    with the db_event.
    '''
    source_events = events.SourceEvent.objects. \
        filter(event=db_event).values_list('source')
    return events.Source.objects.filter(id__in=source_events)


def get_destinations(db_event: events.Event):
    '''
    Returns each of the destination events associated with the db_event
    parameter.

    :param: db_event: The event used to look up destinations.
    :return: a QuerySet containing the destination model instances associated
    with the db_event.
    '''
    destination_events = events.DestinationEvent.objects. \
        filter(event=db_event).values_list('destination')
    return events.Destination.objects.filter(id__in=destination_events)


class EPCISDBProxy:
    '''
    Acts as a proxy between the abstracted database model and
    the EPCIS schema / XML model by converting queries for data into
    EPCPyYes objects.
    '''
    def get_message_by_event_id(self, event_id: str, return_header=True):
        message_id = events.Event.objects.get(
            id = event_id
        ).message_id
        message = headers.Message.objects.get(id = message_id)
        return self.get_full_message(message, return_header)

    def get_full_message(self, message: headers.Message, return_header=True):
        '''
        Returns all of the events and the header in a given message as
        a collection of EPCPyYes class instances.
        :param message: The message to use as the lookup for the header
        and event instances.
        :return: An EPCPyYes.core.v1_2.events.EPCISDocument with all parts
        associated with the given message.
        '''
        document = template_events.EPCISDocument()
        if return_header:
            try:
                # first get the header if there was one
                db_header = headers.SBDH.objects.get(message=message)
                document.header = self._get_header(db_header)
            except SBDH.DoesNotExist:
                logger.debug('There was no document header associated '
                             'with message %s', message)
        # now get the events
        db_events = events.Event.objects.prefetch_related(
            'transformationid_set',
            'errordeclaration_set',
            'quantityelement_set',
            'businesstransaction_set',
            'instancelotmasterdata_set',
            'sourceevent_set__source',
            'destinationevent_set__destination',
        ).filter(message_id=message.id)
        pevents = [self.get_epcis_event(db_event) for db_event in
                   db_events]
        for event in pevents:
            if isinstance(event, pyyes_events.TransformationEvent):
                document.transformation_events.append(event)
            elif isinstance(event, pyyes_events.TransactionEvent):
                document.transaction_events.append(event)
            elif isinstance(event, pyyes_events.ObjectEvent):
                document.object_events.append(event)
            elif isinstance(event, pyyes_events.AggregationEvent):
                document.aggregation_events.append(event)
        return document

    def get_entries_by_parent_identifier(self, identifier: str,
                                         select_for_update=True):
        '''
        Returns a QuerySet of entries based on the incoming identifier
        of the parent.
        :param identifier: The identifier field of the parent Entry.
        :return: A QuerySet of Entry model instances.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            parent_id__identifier=identifier,
            decommissioned=False
        )

    def get_epcs_by_parent_identifier(self, identifier: str,
                                      select_for_update=True):
        '''
        Returns a list of EPCs (identifier values) based on
        the incoming identifier of the parent.
        :param identifier: The identifier field of the parent Entry.
        :return: A list of strings representing the child epcs/identifiers.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            parent_id__identifier=identifier,
            decommissioned=False
        ).values_list('identifier', flat=True)

    def get_events_by_epc(self, epc: str = None, epc_pk: str = None):
        '''
        Returns a list of EPCPyEvents the epc was found in.
        :param epc: The epc to search events for.
        :param epc_pk: The primary key of the epc to search events for.
        :return: A list of EPCPyEvents
        '''
        args = {'identifier': epc} if epc else {'id': epc_pk}
        event_entries = entries.EntryEvent.objects.order_by(
            'event__event_time'
        ).select_related(
            'event'
        ).prefetch_related(
            'event__transformationid_set',
            'event__errordeclaration_set',
            'event__quantityelement_set',
            'event__businesstransaction_set',
            'event__instancelotmasterdata_set',
            'event__sourceevent_set__source',
            'event__destinationevent_set__destination',
        ).filter(**args)
        return [self.get_epcis_event(event_entry.event) for event_entry in
                event_entries]

    def get_epcis_event(self, db_event: events.Event):
        '''
        Takes the raw database event record and converts it to an EPCPyYes
        event.
        :param db_event: The database event instance.
        :return: An EPCPyYes template event of type Object,
        Transformation, Transaction or Aggregation.
        '''
        # look up the type
        if db_event.type == EventTypeChoicesEnum.OBJECT.value:
            ret = self._get_object_event(db_event)
        elif db_event.type == EventTypeChoicesEnum.AGGREGATION.value:
            ret = self._get_aggregation_event(db_event)
        elif db_event.type == EventTypeChoicesEnum.TRANSACTION.value:
            ret = self._get_transaction_event(db_event)
        elif db_event.type == EventTypeChoicesEnum.TRANSFORMATION.value:
            ret = self._get_transformation_event(db_event)
        else:
            ret = None
        # proxy out the call to the specific function
        if ret:
            ret.id = db_event.id
        return ret

    def get_events_by_ilmd(self, name, value):
        '''
        Returns a list of EPCPyYes events by ILMD name value pair.
        Usefull for getting all events for a Lot, etc.
        :param name: The ILMD field name.
        :param value: The ILMD field value.
        :return: A list of EPCPyYes template_event instances.
        '''
        ilmds = events.InstanceLotMasterData.objects.select_related(
            'event'
        ).prefetch_related(
            'event__transformationid_set',
            'event__errordeclaration_set',
            'event__quantityelement_set',
            'event__businesstransaction_set',
            'event__instancelotmasterdata_set',
            'event__sourceevent_set__source',
            'event__destinationevent_set__destination',
        ).filter(
            name=name,
            value=value
        )
        return [self.get_epcis_event(ilmd.event) for ilmd in ilmds]

    def get_event_by_id(self, event_id: str):
        '''
        Looks up an event by it's primary key or event_id value.
        :param event_id: an event identifier that corresponds to one
        of the two values.
        :return: An Event model instance.
        '''
        db_event = events.Event.objects.get(
            Q(id=event_id) | Q(event_id=event_id))
        return self.get_epcis_event(db_event)

    def get_sbdh(self, instance_identifier: str):
        '''
        Use the instance identifier to retrieve a full EPCIS document
        standard business document header.
        :param instance_identifier: The unique id for the document.
        :return: A full EPCPyYes representation of the EPCIS message.
        '''
        try:
            db_header = headers.SBDH.objects.select_related(
                'document_identification',
                'message').prefetch_related('partner_set').get(
                document_identification__instance_identifier=
                instance_identifier)
            header = self._get_header(db_header)
            return header
        except headers.DocumentIdentification.DoesNotExist:
            raise headers.DocumentIdentification.DoesNotExist(
                _('The EPCIS document with instance identifier %s '
                  'could not be found in the database.' % instance_identifier)
            )

    def _get_header(self, db_header):
        '''
        Constucts an EPCPyYes header from the database header supplied.
        :param db_header:
        :return: An EPCPyYes header.
        '''
        if db_header.document_identification:
            # get the document identification data
            header = template_sbdh.StandardBusinessDocumentHeader()
            header.document_identification.instance_identifier = \
                db_header.document_identification.instance_identifier
            header.document_identification.creation_date_and_time = \
                db_header.document_identification.creation_date_and_time
            header.document_identification.standard = \
                db_header.document_identification.standard
            header.document_identification.type_version = \
                db_header.document_identification.type_version
            header.document_identification.multiple_type = \
                db_header.document_identification.multiple_type
            header.document_identification.document_type = \
                db_header.document_identification.document_type
        # get the partners
        header.partners = self.get_partner_list(db_header)
        return header

    def get_partner_list(self, db_header: headers.SBDH):
        '''
        Gets partner list information from the database and returns
        as EPCPyYes partner list.
        :param db_header: The database SBDH model instance.
        :return: A list of EPCPyYes.core.SBDH.sbdh.Partner instances
        '''
        ret = []
        for db_partner in db_header.partner_set.all():
            partner = sbdh.Partner(partner_type=sbdh.PartnerType(
                db_partner.partner_type))
            partner.partner_id = db_partner.identifier
            partner.contact_type_identifier = db_partner.contact_type_identifier
            partner.contact = db_partner.contact
            partner.telephone_number = db_partner.telephone_number
            partner.fax_number = db_partner.fax_number
            partner.email_address = db_partner.email_address
            ret.append(partner)
        return ret

    def get_error_declaration(self, db_event: events.Event,
                              p_event: pyyes_events.EPCISBusinessEvent):
        '''
        Creates a new ErrorDeclaration EPCPyYes instance and appends
        to p_event.
        :param db_event: The database model with the error declaration data.
        :param p_event: The the EPCPyYes business event to append the
        declaration to.
        '''
        # force eval of the queryset
        ed_list = list(db_event.errordeclaration_set.all())
        for ed in ed_list:
            error_declaration = pyyes_events.ErrorDeclaration()
            # get the event ids
            if ed.corrective_event_ids:
                error_declaration.corrective_event_ids = \
                    ed.corrective_event_ids.split(',')
            else:
                ed.corrective_event_ids = []
            # get the reason
            error_declaration.reason = ed.reason
            # get the declaration time and convert to iso string
            error_declaration.declaration_time = ed.declaration_time.isoformat()
            # append the ed to the epcpyyes event
            p_event.error_declaration = error_declaration

    def get_business_transactions(self, db_event: events.Event):
        '''
        Gets the business transaction info from an event and
        returns an EPCPyYes BusinessTransaction.
        :param db_event: The database event.
        :return: An EPCPyYes BusinessTransaction.
        '''
        bts = []
        bt_list = list(db_event.businesstransaction_set.all())
        for db_bt in bt_list:
            bt = pyyes_events.BusinessTransaction(db_bt.biz_transaction)
            bt.type = db_bt.type
            bts.append(bt)
        return bts

    def get_business_event(self, db_event: events.Event,
                           p_event: pyyes_events.EPCISBusinessEvent):
        '''
        Gets all the fields for a business event (the base
        event abstract model)
        :param db_event: The event
        :return:
        '''
        logger.debug('Getting business event data for db event %s',
                     str(db_event.id))
        p_event.biz_location = db_event.biz_location
        p_event.read_point = db_event.read_point
        p_event.disposition = db_event.disposition
        p_event.biz_step = db_event.biz_step
        p_event.action = db_event.action
        self.get_base_epcis_event(db_event, p_event)
        p_event.business_transaction_list = self.get_business_transactions(
            db_event
        )
        p_event.source_list = self.get_source_list(db_event)
        p_event.destination_list = self.get_destination_list(db_event)
        return p_event

    def get_base_epcis_event(self, db_event, p_event):
        '''
        All of the EPCPyYes events share a common base class.  This
        function pulls the data relative to the db_event out of the
        database and populates the base EPCPyYes EPCISEvent class
        fields.
        :param db_event: The database model class instance for the event.
        :param p_event: The EPCPyYes event that inherits from the
        base EPCISEvent class.
        '''
        logger.debug('Getting base epcis data for db event %s',
                     str(db_event.id))
        self.get_error_declaration(db_event, p_event)
        p_event.event_time = db_event.event_time.isoformat()
        p_event.event_timezone_offset = db_event.event_timezone_offset
        p_event.record_time = db_event.record_time.isoformat()
        p_event.event_id = db_event.event_id

    def get_source_list(self, db_event: events.Event):
        '''
        Pulls each source out of the backend datastore and
        serializes to a list of EPCPyYes Source instances.
        :param db_event: The database Event model.
        :return: A list of EPCPyYes.core.v1_2.events.Source instances.
        '''
        ret = []
        se_list = events.SourceEvent.objects.select_related('source').filter(
            event=db_event
        )
        for se in se_list:
            source = pyyes_events.Source(
                source_type=se.source.type,
                source=se.source.source
            )
            ret.append(source)
        return ret

    def get_destination_list(self, db_event: events.Event):
        '''
        Pulls each destination out of the backend datastore and
        serializes to a list of EPCPyYes destination instances.
        :param db_event: The database Event model.
        :return: A list of EPCPyYes.core.v1_2.events.destination instances.
        '''
        ret = []
        dest_list = events.DestinationEvent.objects.select_related(
            'destination').filter(
            event=db_event
        )
        for dest in dest_list:
            destination = pyyes_events.Destination(
                destination_type=dest.destination.type,
                destination=dest.destination.destination
            )
            ret.append(destination)
        return ret

    def get_parent_epc(self, db_event: events.Event):
        '''
        For a given agg or transaction event, will get the parent epc.
        :param db_event: The event to retrieve the parent for.
        :return: A string representing the epc.
        '''
        try:
            ee = entries.EntryEvent.objects.select_related('entry').get(
                event=db_event,
                is_parent=True
            )
            return ee.entry.identifier
        except entries.EntryEvent.DoesNotExist:
            logger.info('No parent for event %s', db_event.id)

    def get_epc_list(
        self, db_event: events.Event, is_parent=False, output=False
    ):
        '''
        Returns all of the EPCs for a given Event.
        :param db_event: The event to look up the EPCs (Entries) for.
        :param is_parent: Whether or not to return any parent ids with
        the result.
        :param output: Whether or not to look for output epc values- only
        for use with transformation events.
        :return: A list of EPCs.
        '''
        ee = entries.EntryEvent.objects.select_related('entry').filter(
            event=db_event,
            is_parent=is_parent,
            output=output
        )
        return [e.entry.identifier for e in ee]

    def get_input_epc_list(self, db_event: events.Event):
        '''
        Just a helper function to make the code more clear when working
        with transformation events.
        :param xform_event: The transformation event to get the input epcs for.
        :return: A list of epc (string) values.
        '''
        return self.get_epc_list(db_event)

    def get_output_epc_list(self, db_event: events.Event):
        '''
        Just a helper function to make the code more clear when working
        with transformation events.
        :param xform_event: The transformation event to get the output
        epcs for.
        :return: A list of epc (string) values.
        '''
        return self.get_epc_list(db_event, output=True)

    def get_quantity_list(self, db_event: events.Event, is_output=False):
        '''
        Retrieves all of the Quantity data for a given Event.
        :param db_event: The Event go get the quantity data for.
        :return:
        '''
        qes = events.QuantityElement.objects.filter(
            event=db_event,
            is_output=is_output
        )
        return [
            pyyes_events.QuantityElement(
                qe.epc_class,
                quantity=qe.quantity,
                uom=qe.uom
            )
            for qe in qes
        ]

    def get_ilmd(self, db_event: events.Event):
        '''
        Gets all of the ILMD entries out of the backend relative to the
        db_event parameter and returns EPCPyYes ILMD class instances.
        :param db_event: The event to find ILMD data for.
        :return: A list of EPCPyYes ILMD class instances.
        '''
        ilmds = events.InstanceLotMasterData.objects.filter(event=db_event)
        return [
            InstanceLotMasterDataAttribute(
                name=ilmd.name,
                value=ilmd.value
            )
            for ilmd in ilmds
        ]

    def _get_object_event(self, db_event: events.Event):
        '''
        Reconstructs a full object event based on the database
        model instance passed in.
        :param db_event: The object event event.Event instance.
        :return: An EPCPyYes template event ObjectEvent.
        '''
        o_event = template_events.ObjectEvent()
        self.get_business_event(db_event, o_event)
        o_event.epc_list = self.get_epc_list(db_event)
        o_event.quantity_list = self.get_quantity_list(db_event)
        o_event.ilmd = self.get_ilmd(db_event)
        return o_event

    def _get_aggregation_event(self, db_event: events.Event):
        '''
        Reconstructs a full aggregation event based on the database
        model instance passed in.
        :param db_event: The aggregation event event.Event instance.
        :return: An EPCPyYes template event AggregationEvent.
        '''
        agg_event = template_events.AggregationEvent()
        self.get_business_event(db_event, agg_event)
        agg_event.child_quantity_list = self.get_quantity_list(db_event)
        agg_event.child_epcs = self.get_epc_list(db_event)
        agg_event.parent_id = self.get_parent_epc(db_event)
        return agg_event

    def _get_transaction_event(self, db_event: events.Event):
        '''
        Reconstructs a full transaction event based on the database
        model instance passed in.
        :param db_event: The transaction event event.Event instance.
        :return: An EPCPyYes template event TransactionEvent.
        '''
        xact_event = template_events.TransactionEvent()
        self.get_business_event(db_event, xact_event)
        xact_event.epc_list = self.get_epc_list(db_event)
        xact_event.parent_id = self.get_parent_epc(db_event)
        xact_event.quantity_list = self.get_quantity_list(db_event)
        return xact_event

    def _get_transformation_event(self, db_event: events.Event):
        '''
        Reconstructs a full transformation event from the data
        in the database relative the the passed in db_event parameter.
        :param db_event: The event to use to pull the relevant data out
        of the database.
        :return: An EPCPyYes template_event TransformationEvent.
        '''
        xform_event = template_events.TransformationEvent()
        # get the basic EPCISEvent values
        self.get_base_epcis_event(db_event, xform_event)
        logger.debug('Handling a transformation event')
        xform_event.input_epc_list = self.get_input_epc_list(db_event)
        xform_event.output_epc_list = self.get_output_epc_list(db_event)
        xform_event.input_quantity_list = self.get_quantity_list(db_event)
        xform_event.output_quantity_list = self.get_quantity_list(
            db_event, is_output=True
        )
        xform_event.destination_list = self.get_destination_list(db_event)
        xform_event.source_list = self.get_source_list(db_event)
        xform_event.ilmd = self.get_ilmd(db_event)
        xform_event.business_transaction_list = self.get_business_transactions(
            db_event
        )
        xform_event.biz_location = db_event.biz_location
        xform_event.read_point = db_event.read_point
        xform_event.disposition = db_event.disposition
        xform_event.biz_step = db_event.biz_step
        xform_event.transformation_id = self.get_transformation_id(db_event)
        return xform_event

    def get_transformation_id(self, db_event: events.Event):
        '''
        Looks up the transformation id for a given transformation event.
        :param db_event: The Event model instance to use for the lookup.
        :return: The transformation id as a string.
        '''
        try:
            tid = events.TransformationID.objects.get(event=db_event)
            return tid.identifier
        except events.TransformationID.DoesNotExist:
            logger.debug('No transformation id was found for event %s',
                         db_event.id)

    def get_entries_by_parent(self, parent_entry: entries.Entry,
                              select_for_update=True):
        '''
        Returns a queryset containing all child Entries for a parent level
        entry.
        :param parent_entry: The parent to retrieve the children for.
        :param select_for_update: Whether or not the returned QuerySet contains
        model instances selected for update by the database.
        :return: A queryset of Entry model instances reflecting the children
        of the parent_entry.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            parent_id__identifier=parent_entry,
            decommissioned=False
        )

    def get_entry_child_parents(self, parent_entry: entries.Entry,
                                select_for_update=True):
        '''
        Returns a queryset containing all child Entries for a parent level
        entry where the children are also parents.
        :param parent_entry: The parent to retrieve the children for.
        :param select_for_update: Whether or not the returned QuerySet contains
        model instances selected for update by the database.
        :return: A queryset of Entry model instances reflecting the children
        of the parent_entry that are also parents.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            parent_id__identifier=parent_entry,
            is_parent=True,
            decommissioned=False
        )

    def get_entries_by_top(self, top_entry: entries.Entry,
                           select_for_update=True):
        '''
        Returns all entries that are under a top_entry.
        :param top_entry: The top-level Entry instance.
        :param select_for_update: Whether or not the returned QuerySet contains
        model instances selected for update by the database.
        :return: A QuerySet of Entry instances.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            top_id=top_entry,
            decommissioned=False
        )

    def get_entries_by_tops(self, top_entries: EntryList,
                            select_for_update=True):
        '''
        Returns all entries that are under the list of top entries.
        :param top_entries: The top-level Entry instances.
        :param select_for_update: Whether or not the returned QuerySet contains
        model instances selected for update by the database.
        :return: A QuerySet of Entry instances.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            top_id__in=top_entries,
            decommissioned=False
        )

    def _update_or_filter(self, select_for_update):
        '''
        Returns a function pointer to select for update filter or the
        regular filter function on the QuerySet of the Entry database
        model.
        :param select_for_update: Whether or not any Entries are selected
        for update by the database.
        :return:
        '''
        if select_for_update:
            func = entries.Entry.objects.select_for_update().filter
        else:
            func = entries.Entry.objects.filter
        return func

    def get_entries_by_parents(self, parents: EntryList,
                               select_for_update=True):
        '''
        Returns a list of entries that are children if the inbound parent
        list.
        :param parents: The parent ids to find children for.
        :param select_for_update: Whether or not the returned QuerySet contains
        model instance selected for update by the database.
        :return: A QuerySet of Entry instances.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            parent_id__in=parents,
            decommissioned=False
        )

    def get_parent_entries(self, epcs: list, select_for_update=True):
        '''
        Out of a list of EPCs, will return any that are parents in a QuerySet
        for update.
        :param epcs: A list of EPCs to inspect for is_parent=True
        :param select_for_update: Whether or not to select any parent entries
        from the database for update in a transaction.
        :return: Any EPCs that have an is_parent value of True and are not
        decommissioned.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            is_parent=True,
            decommissioned=False,
            identifier__in=epcs
        )

    def get_top_entries(self, epcs: list, select_for_update=True):
        '''
        Out of a list of EPCs, will return any that are top-level entries
        in a QuerySet for update.
        :param epcs: A list of EPCs to inspect for is_parent=True, top_id=None
        and parent_id=None (and not decommissioned)
        :param select_for_update: Whether or not to select any parent entries
        from the database for update in a transaction.
        :return: Any EPCs that have an is_parent value of True and are not
        decommissioned.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            identifier__in=epcs,
            is_parent=True,
            parent_id=None,
            top_id=None,
            decommissioned=False,
        )

    def get_entries_by_epcs(self, epcs: list, select_for_update=True):
        '''
        Returns a queryset of Entry model instances that have identifiers
        correlating to values in the `epcs` list.
        :param epcs: The epcs to search for.
        :param select_for_update: Whether or not to select any entries
        from the database for update in a transaction. Default=True
        :return: A QuerySet of Entries.
        '''
        func = self._update_or_filter(select_for_update)
        return func(
            identifier__in=epcs,
            decommissioned=False
        )

    def get_entries_by_event(self, db_event: events.Event):
        '''
        Returns a list all of the entries (serial numbers) for each event.
        :param db_event: The event to retrieve the entries for.
        :return: A list of entry model instances.
        '''
        ee = entries.EntryEvent.objects.select_related('entry').filter(
            event=db_event
        )
        return [e.entry for e in ee]

    def get_events_by_entry_list(self, entry_list: EntryList,
                                 event_type: str = None):
        '''
        Based on an inbound list of entries, will find all the events
        associated and return them as EPCPyYes events.
        :param entry_list: A list of Entry model instances.
        :param event_type: Specify a specific event type if you are interested
        in returning just Aggregation events for example.  The event type
        can be specified by using the
        `quartet_epcis.models.choices.EventTypeChoicesEnum` enumeration.
        For example `event_type = EventTypeChoicesEnum.Aggregation.value`
        :return: A QuerySet of events.
        '''
        kwargs = {
            'entry__in': entry_list,
        }
        if event_type: kwargs['event_type'] = event_type
        db_entry_events = entries.EntryEvent.objects.filter(
            **kwargs
        ).prefetch_related(
            'event'
        ).values_list(
            'event', flat=True
        ).distinct()
        db_events = events.Event.objects.filter(
            id__in=db_entry_events
        )
        return [self.get_epcis_event(db_event) for db_event in db_events]

    def get_events_by_entry_identifer(self, entry_identifier: str):
        '''
        Pulls all of the events out of the database based associated
        with the Entry.
        :param entry_identifier: The entry to use to lookup the event.
        :return: An models.Event instance.
        '''
        # get the entry_events
        ret = []
        events = list(
            entries.EntryEvent.objects.select_related('event').only(
                'event').filter(
                entry__identifier=entry_identifier
            ))
        for event in events:
            ret.append(event)
        return ret

    def get_aggregation_parents_by_epcs(self, epcs: list):
        """
        Will return a dictionary of EPCPyYes aggregation events
        based on the input
        list.  The assumption is that the list contains a number of parent
        epc values (the identifier field in the
        `quartet_epcis.models.entries.Entry` model.  Any children of the
        parents will also be returned.  Basically returns everything except
        the bottom level entries in a hierarchy.
        :param epcs: A list of epc urns or identifiers that can be used
        to lookup an Entry by its indentifier field.
        :return: A dictionary of
        `EPCPyYes.core.v1_2.template_events.AggregationEvents` with the key
        as the identifier (urn) and the value being the event.
        """
        # get all the epcs including children
        # first get the entries
        collected_entries = {}
        db_entries = self.get_entries_by_epcs(epcs, select_for_update=False)
        for db_entry in db_entries:
            # add this to the collection...
            # if there is no top_id and it is a parent that means it's a top
            lower_entries = self.get_entry_child_parents(db_entry, select_for_update=False)
            collected_entries[db_entry.identifier] = db_entry

            for lower_entry in lower_entries:
                # add the children to the connection
                collected_entries[
                    lower_entry.identifier] = lower_entry
                # recursively call this function for the lower entry is also
                # a parent
                if lower_entry.is_parent:
                    child_entries = self.get_aggregation_parents_by_epcs(
                        [lower_entry]
                    )
                    # merge the dictionaries
                    collected_entries = {**collected_entries, **child_entries}
        return collected_entries

    def get_aggregation_events_by_epcs(self, epcs: list):
        '''
        When supplied with a list of top level ids or parent ids,
        will get a list of EPCPyYes aggregation events corresponding to their
        aggregation event history.  Useful for transmitting full histories
        of packing, repacking, etc.
        :param eps: The list of parents.
        :return: A list of EPCPyYes template_event AggregationEvent instances.
        '''
        ret = []
        top_entries = self.get_aggregation_parents_by_epcs(epcs)
        # now that we have a comprehensive list of parent entries,
        # we can get all of the aggregation events associated
        db_events = entries.EntryEvent.objects.select_related(
            'event'
        ).filter(
            entry__in=list(top_entries.values()),
            event_type=EventTypeChoicesEnum.AGGREGATION.value,
            is_parent=True
        ).values('event').distinct()
        db_events = events.Event.objects.filter(
            id__in=db_events
        )
        for db_event in db_events:
            ret.append(self._get_aggregation_event(db_event))
        return ret

    def get_object_events_by_epcs(self, epcs: list):
        '''
        When supplied with a list of epcs,
        will get a list of EPCPyYes object events corresponding to their
        event history.  Useful for transmitting full histories
        of packing, repacking, etc.
        :param eps: The list of parents.
        :return: A list of EPCPyYes template_event AggregationEvent instances.
        '''
        ret = []
        db_entries = self.get_entries_by_epcs(epcs)
        db_events = entries.EntryEvent.objects.select_related(
            'event'
        ).filter(
            entry__in=list(db_entries),
            event_type=EventTypeChoicesEnum.OBJECT.value,
        ).values('event').distinct()
        db_events = events.Event.objects.filter(
            id__in=db_events
        )
        for db_event in db_events:
            ret.append(self._get_object_event(db_event))
        return ret

    def _get_event_entries(self, db_event: events.Event):
        '''
        Returns all of the entries for a given event.
        :param db_event:
        :return: A list of entries.
        '''
        ret = []
        eevs = list(entries.EntryEvent.objects.select_related('entry').only(
            'event').filter(
            event=db_event))
        for eev in eevs:
            ret.append(eev.entry)
        return ret

    def _convert_event_to_pyyes(self, db_event: events.Event):
        '''
        Converts a database event into an EPCPyYes event.
        :param db_event: The database event
        :return: An EPCPyYes.core.v1_2 template event instance.
        '''
        # get the event's serial numbers
        entries = self._get_event_entries(db_event)
        # get the event's basic stuff
        if db_event.type == EventTypeChoicesEnum.OBJECT.value:
            event = self._get_object_event(db_event)
        elif db_event.type == EventTypeChoicesEnum.AGGREGATION.value:
            event = self._get_aggregation_event(db_event)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from rest_framework.serializers import ModelSerializer
from quartet_epcis.models import entries, events, headers


class EntrySerializer(ModelSerializer):
    class Meta:
        model = entries.Entry
        fields = '__all__'


class EntryEventSerialzier(ModelSerializer):
    class Meta:
        model = entries.EntryEvent
        fields = '__all__'


class EventSerializer(ModelSerializer):
    class Meta:
        model = events.Event
        fields = '__all__'


class TransformationIDSerializer(ModelSerializer):
    class Meta:
        model = events.TransformationID
        fields = '__all__'


class ErrorDeclarationSerializer(ModelSerializer):
    class Meta:
        model = events.ErrorDeclaration
        fields = '__all__'


class QuantityElementSerializer(ModelSerializer):
    class Meta:
        model = events.QuantityElement
        fields = '__all__'


class BusinessTransactionSerializer(ModelSerializer):
    class Meta:
        model = events.BusinessTransaction
        fields = '__all__'


class InstanceLotMasterDataSerializer(ModelSerializer):
    class Meta:
        model = events.InstanceLotMasterData
        fields = '__all__'


class SourceSerializer(ModelSerializer):
    class Meta:
        model = events.Source
        fields = '__all__'


class SourceEventSerializer(ModelSerializer):
    class Meta:
        model = events.SourceEvent
        fields = '__all__'


class DestinationSerializer(ModelSerializer):
    class Meta:
        model = events.Destination
        fields = '__all__'


class DestinationEventSerializer(ModelSerializer):
    class Meta:
        model = events.DestinationEvent
        fields = '__all__'


class MessageSerializer(ModelSerializer):
    class Meta:
        model = headers.Message
        fields = '__all__'



# Generated by Django 2.0 on 2018-10-31 19:20

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('quartet_epcis', '0002_auto_20180718_1528'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='event',
            options={'ordering': ['-event_time'], 'verbose_name': 'Event', 'verbose_name_plural': 'Events'},
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-18 15:28

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import quartet_epcis.models.abstractmodels
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('quartet_epcis', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='BusinessTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('biz_transaction', models.CharField(help_text='The business transaction.', max_length=200, verbose_name='Business Transaction')),
                ('type', models.CharField(help_text='The type of business transaction.', max_length=200, null=True, verbose_name='Type')),
            ],
            options={
                'verbose_name': 'Business Transaction',
                'verbose_name_plural': 'Business Transactions',
            },
        ),
        migrations.CreateModel(
            name='Destination',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='Unique ID', primary_key=True, serialize=False, verbose_name='Unique ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this record was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='When this record was last modified.', verbose_name='Modified')),
                ('type', models.CharField(help_text='The source type.', max_length=150, verbose_name='Type')),
                ('destination', models.CharField(help_text='The Destination identifier.', max_length=150, verbose_name='Destination')),
            ],
            options={
                'verbose_name': 'Destination',
                'verbose_name_plural': 'Destinations',
            },
        ),
        migrations.CreateModel(
            name='DestinationEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('destination', models.ForeignKey(help_text='A destination within the event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Destination', verbose_name='Destination')),
            ],
        ),
        migrations.CreateModel(
            name='DocumentIdentification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('standard', models.CharField(default='EPCglobal', help_text='The originator of the standard that the following document falls under.  Default is EPCglobal.', max_length=20, verbose_name='Standard')),
                ('type_version', models.CharField(default='1.0', help_text='Descriptor which contains versioning information or number of the standard that defines the documentwhich is specified in the ’Type’ data element, e.g. values could be ‘1.3’ or ‘D.96A’, etc.', max_length=10, verbose_name='Version')),
                ('instance_identifier', models.CharField(db_index=True, help_text='Descriptor which contains reference information which uniquely identifies this instance of the SBD between the sender and the receiver.', max_length=100, verbose_name='Instance Identifier')),
                ('document_type', models.CharField(choices=[('Events', 'Events'), ('MasterData', 'MasterData'), ('QueryControl-Request', 'QueryControl-Request'), ('QueryControl-Response', 'QueryControl-Response'), ('QueryCallback', 'QueryCallback'), ('Query', 'Query')], help_text='A logical indicator representing the type of Business Data being sent or the named type of business data.', max_length=25, verbose_name='Type')),
                ('multiple_type', models.NullBooleanField(default=False, help_text='A flag to indicate that there is more than one type of Document in the instance.', verbose_name='Multiple Type')),
                ('creation_date_and_time', models.CharField(help_text='Descriptor which contains date and time of SBDH/document creation.', max_length=35, null=True, verbose_name='Creation Date and Time')),
            ],
            options={
                'verbose_name': 'Document Identification',
                'verbose_name_plural': 'Document Identifications',
            },
        ),
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='Unique ID', primary_key=True, serialize=False, verbose_name='Unique ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this record was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='When this record was last modified.', verbose_name='Modified')),
                ('identifier', models.CharField(db_index=True, help_text='The primary unique id for the entry.', max_length=150, unique=True, verbose_name='EPC URN')),
                ('last_event_time', models.DateTimeField(help_text='The time of the event that last affected the status ofthis entry.', null=True, verbose_name='Last Event Time')),
                ('last_disposition', models.CharField(help_text='The business condition of the objects associated with the EPCs, presumed to hold true until contradicted by a subsequent event..', max_length=150, null=True, verbose_name='Last Disposition')),
                ('last_aggregation_event_time', models.DateTimeField(help_text='The time of the event that last affected the status ofthis entries hierarchical relation to other entries.', null=True, verbose_name='Last Aggregation Event Time')),
                ('last_aggregation_event_action', models.CharField(choices=[('ADD', 'Add'), ('OBSERVE', 'Observe'), ('DELETE', 'Delete')], help_text='The action (ADD or DELETE) of the last aggregation event that affected this entry.  Observation events are not noted.', max_length=10, null=True, verbose_name='Last Aggregation Action')),
                ('is_parent', models.BooleanField(default=False, help_text='True if this entry is a parent in any hierarchies. Falseif not.', verbose_name='Is Parent')),
                ('decommissioned', models.BooleanField(default=False, help_text='Whether or not the entry has been decommissioned.  Oncean entry is decommissioned, it can no longer take placein business processes.', verbose_name='Decommissioned')),
            ],
            options={
                'verbose_name': 'Entry',
                'verbose_name_plural': 'Entries',
                'ordering': ['created'],
            },
        ),
        migrations.CreateModel(
            name='EntryEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event_type', models.CharField(choices=[('ag', 'Aggregation'), ('ob', 'Object'), ('tx', 'Transaction'), ('tf', 'Transformation')], help_text='The type of event (Aggregation, Object, Transaction or Transformation.', max_length=3, verbose_name='Event Type')),
                ('event_time', models.DateTimeField(db_index=True, help_text="The Event's eventTime.", verbose_name='Event Time')),
                ('identifier', models.CharField(db_index=True, help_text='A redundant entry ID entry for fast event composition.', max_length=150, verbose_name='EPC URN')),
                ('is_parent', models.BooleanField(default=False, help_text="Whether or not this entry was the parent of it's constituent event.", verbose_name='Is Event Parent')),
                ('output', models.BooleanField(default=False, help_text='Whether or not the entry was the output of a Transformation event.', verbose_name='Transformation Output')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this record was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='When this record was last modified.', verbose_name='Modified')),
                ('entry', models.ForeignKey(help_text='The Unique ID of the Entry', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Entry', verbose_name='Entry ID')),
            ],
            options={
                'verbose_name': 'Entry Event Record',
                'verbose_name_plural': 'Entry Event Records',
            },
        ),
        migrations.CreateModel(
            name='ErrorDeclaration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('declaration_time', models.DateTimeField(default=django.utils.timezone.now, help_text='The time at which the error was declared.', verbose_name='Declaration Time')),
                ('reason', models.CharField(help_text='The reason for the error.', max_length=150, null=True, verbose_name='Reason')),
                ('corrective_event_ids', models.TextField(help_text='A delimited list of EPCIS event ids.', null=True, verbose_name='Corrective Event IDs')),
            ],
            options={
                'verbose_name': 'Error Declaration',
                'verbose_name_plural': 'Error Declarations',
            },
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='Unique ID', primary_key=True, serialize=False, verbose_name='Unique ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this record was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='When this record was last modified.', verbose_name='Modified')),
                ('event_time', models.DateTimeField(db_index=True, editable=False, help_text='The date and time at which the EPCIS Capturing Application asserts the event occurred.', verbose_name='Event Time')),
                ('event_timezone_offset', models.CharField(default='+00:00', help_text='The time zone offset in effect at the time and place the event occurred, expressed as an offset from UTC', max_length=6, null=True, verbose_name='Event Timezone Offset')),
                ('record_time', models.DateTimeField(default=django.utils.timezone.now, help_text='The date and time at which this event was recorded by an EPCIS Repository.', null=True, verbose_name='Record Time')),
                ('event_id', models.CharField(db_index=True, default=quartet_epcis.models.abstractmodels.haikunate, help_text='An identifier for this event as specified by the capturing application, globally unique across all events other than error declarations. Not to be confused with the unique id/primary key for events within a database.', max_length=150, null=True, verbose_name='Event ID')),
                ('action', models.CharField(choices=[('ADD', 'Add'), ('OBSERVE', 'Observe'), ('DELETE', 'Delete')], help_text='How this event relates to the lifecycle of the EPCs named in this event.', max_length=10, verbose_name='Action')),
                ('biz_step', models.CharField(help_text='The business step of which this event was a part.', max_length=150, null=True, verbose_name='Business Step')),
                ('disposition', models.CharField(help_text='The business condition of the objects associated with the EPCs, presumed to hold true until contradicted by a subsequent event..', max_length=150, null=True, verbose_name='Disposition')),
                ('read_point', models.CharField(help_text='The read point at which the event took place.', max_length=150, null=True, verbose_name='Read Point')),
                ('biz_location', models.CharField(help_text='The business location where the objects associated with the EPCs may be found, until contradicted by a subsequent event.', max_length=150, null=True, verbose_name='Business Location')),
                ('type', models.CharField(choices=[('ag', 'Aggregation'), ('ob', 'Object'), ('tx', 'Transaction'), ('tf', 'Transformation')], help_text='The type of event.', max_length=2, verbose_name='Event Type')),
                ('message_id', models.CharField(help_text='The unique id of the originating message.', max_length=100, verbose_name='Message ID')),
            ],
            options={
                'verbose_name': 'Event',
                'verbose_name_plural': 'Events',
                'ordering': ['event_time'],
            },
        ),
        migrations.CreateModel(
            name='InstanceLotMasterData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of the ILMD entry.', max_length=150, verbose_name='Name')),
                ('value', models.CharField(help_text='The value of the ILMD entry.', max_length=255, verbose_name='Value')),
                ('event', models.ForeignKey(help_text='The source event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event')),
            ],
            options={
                'verbose_name': 'ILMD Entry',
                'verbose_name_plural': 'ILMD Entries',
            },
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='Partner',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('partner_type', models.CharField(choices=[('Sender', 'Sender'), ('Receiver', 'Receiver')], help_text='The type of partner.  Either Sender or Receiver.', max_length=20, verbose_name='Partner Type')),
                ('authority', models.CharField(default='SGLN', help_text='The identifying authority/format for the identification field. Default is SGLN.', max_length=20, null=True, verbose_name='authority')),
                ('identifier', models.CharField(help_text='An identifier that is in line with the authority specified in the authority field.  Typically an SGLN URN value.', max_length=100, null=True, verbose_name='Identifier')),
                ('contact', models.CharField(help_text='The contact/name info.', max_length=50, null=True, verbose_name='Contact')),
                ('email_address', models.EmailField(help_text='The email address for the partner.', max_length=100, null=True, verbose_name='Email Address')),
                ('fax_number', models.CharField(help_text='Fax number.', max_length=20, null=True, verbose_name='Fax Number')),
                ('telephone_number', models.CharField(help_text='Telephone number for the partner.', max_length=20, null=True, verbose_name='Telephone number.')),
                ('contact_type_identifier', models.CharField(help_text='The type of contact- for example, ', max_length=40, null=True, verbose_name='Role of the contact in this business process.')),
            ],
            options={
                'verbose_name': 'Partner',
                'verbose_name_plural': 'Partners',
            },
        ),
        migrations.CreateModel(
            name='QuantityElement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('epc_class', models.CharField(help_text='The EPC class.', max_length=200, verbose_name='EPC Class')),
                ('quantity', models.FloatField(help_text='The Quantity value.', verbose_name='Quantity')),
                ('uom', models.CharField(help_text='The unit of measure relative to the quantity.', max_length=150, null=True, verbose_name='Unit of Measure (UOM)')),
                ('is_output', models.BooleanField(default=False, help_text='True if this quantity element was provided as the output as part of a transformation event.', verbose_name='Is Output')),
                ('event', models.ForeignKey(help_text='The source event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event')),
            ],
            options={
                'verbose_name': 'Quantity Element',
                'verbose_name_plural': 'Quantity Elements',
            },
        ),
        migrations.CreateModel(
            name='SBDH',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('header_version', models.CharField(default='1.0', help_text='Descriptor which contains version information for the SBDH.  Default is 1.0', max_length=10, verbose_name='Header Version')),
                ('document_identification', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.DocumentIdentification')),
                ('message', models.ForeignKey(help_text='The message this header was associated with.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Message', verbose_name='Message')),
            ],
            options={
                'verbose_name': 'SBDH',
                'verbose_name_plural': 'SBDHs',
            },
        ),
        migrations.CreateModel(
            name='Source',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='Unique ID', primary_key=True, serialize=False, verbose_name='Unique ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this record was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='When this record was last modified.', verbose_name='Modified')),
                ('type', models.CharField(help_text='The source type.', max_length=150, verbose_name='Type')),
                ('source', models.CharField(help_text='The source identifier.', max_length=150, verbose_name='Source')),
            ],
            options={
                'verbose_name': 'Source',
                'verbose_name_plural': 'Sources',
            },
        ),
        migrations.CreateModel(
            name='SourceEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event', models.ForeignKey(help_text='The event within which the source was reported.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event')),
                ('source', models.ForeignKey(help_text='A source within the event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Source', verbose_name='Source')),
            ],
        ),
        migrations.CreateModel(
            name='TransformationID',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, help_text='Unique ID', primary_key=True, serialize=False, verbose_name='Unique ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this record was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='When this record was last modified.', verbose_name='Modified')),
                ('identifier', models.CharField(help_text='The Transformation event ID.', max_length=150, verbose_name='TransformationID')),
                ('event', models.ForeignKey(help_text='The source event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event')),
            ],
            options={
                'verbose_name': 'Transformation ID',
                'verbose_name_plural': 'Transformation IDs',
            },
        ),
        migrations.AddField(
            model_name='partner',
            name='header',
            field=models.ForeignKey(help_text='The related SBDH.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.SBDH', verbose_name='SBDH'),
        ),
        migrations.AddField(
            model_name='errordeclaration',
            name='event',
            field=models.ForeignKey(help_text='The source event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event'),
        ),
        migrations.AddField(
            model_name='entryevent',
            name='event',
            field=models.ForeignKey(help_text='The UUID of the event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event ID'),
        ),
        migrations.AddField(
            model_name='entry',
            name='last_aggregation_event',
            field=models.ForeignKey(help_text='Used mostly for internal tracking and performance duringparsing.  This tracks the last aggregation event thataffected the entry.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='last_agg_event', to='quartet_epcis.Event', verbose_name='Last Aggregation Event'),
        ),
        migrations.AddField(
            model_name='entry',
            name='last_event',
            field=models.ForeignKey(help_text='The last event to affect the status of this entry.', null=True, on_delete=django.db.models.deletion.PROTECT, to='quartet_epcis.Event', verbose_name='Last Event'),
        ),
        migrations.AddField(
            model_name='entry',
            name='parent_id',
            field=models.ForeignKey(help_text='The parent of this identifier (if any).', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='parent_identifier', to='quartet_epcis.Entry', verbose_name='Parent ID'),
        ),
        migrations.AddField(
            model_name='entry',
            name='top_id',
            field=models.ForeignKey(help_text='The top level id (if any).', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='top_identifier', to='quartet_epcis.Entry', verbose_name='Top ID'),
        ),
        migrations.AddField(
            model_name='destinationevent',
            name='event',
            field=models.ForeignKey(help_text='The event within which the destination was reported.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event'),
        ),
        migrations.AddField(
            model_name='businesstransaction',
            name='event',
            field=models.ForeignKey(help_text='The source event.', on_delete=django.db.models.deletion.CASCADE, to='quartet_epcis.Event', verbose_name='Event'),
        ),
        migrations.AlterIndexTogether(
            name='entryevent',
            index_together={('event', 'entry')},
        ),
    ]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import io

from enum import Enum
from quartet_capture import models
from quartet_capture.rules import Step as RuleStep
from quartet_capture.rules import RuleContext
from quartet_epcis.parsing.parser import QuartetParser
from quartet_epcis.parsing.business_parser import BusinessEPCISParser
from quartet_epcis.parsing.json import JSONParser
from django.core.files.base import File
from quartet_capture.models import Rule, Step, StepParameter
from django.utils.translation import gettext as _


class ContextKeys(Enum):
    """
    Contains context keys that steps within this module can / will place
    on the rule context during processing.

    EPCIS_MESSAGE_ID_KEY
    --------------------
    This is the message id that identifies all of the events, headers and
    entries that are parsed when the EPCISParsingStep is called.  When
    the EPCIS parser saves entries, headers and event model instances, it
    associates them all with an inbound message id by creating a
    Message model instance and associating all of these other model instance
    with that message.id property. Once the parsing step is complete it
    will place the message id on the context using this key.
    """
    EPCIS_MESSAGE_ID_KEY = 'MESSAGE_ID'


def create_rule():
    '''
    Creates the default rule.  Used in data-migrations.
    :return: None
    '''
    rule = Rule.objects.create(
        name=_('EPCIS'),
        description=_('Will capture and parse all properly formed inbound '
                      'EPCIS messagess.  Loose or strict enforcement can '
                      'be controlled via step parameters.'),
    )
    step = Step.objects.create(
        name=_('Parse XML'),
        description=_('Parse EPCIS data and save to database. To set loose '
                      'enforcement (capture all messages) change the '
                      '"LooseEnforcement" step parameter to have a '
                      'value of True.'),
        step_class='quartet_epcis.parsing.steps.EPCISParsingStep',
        order=1,
        rule=rule
    )
    StepParameter.objects.create(
        step=step,
        name='LooseEnforcement',
        value='False',
        description=_('If set to true, QU4RTET will capture all properly '
                      'formed EPCIS messages regardless of business context.')
    )


class EPCISParsingStep(RuleStep):
    '''
    Calls the EPCIS parser as a rules.Step that can be used in the
    quartet_capture rule engine.  Puts the Message id on the rule context as
    EPCIS_MESSAGE_ID.
    '''

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        # check to see which parser to use if loose enforcement, then
        # use the quartet parser which just captures messages without
        # trying to enforce any business rules (good for testing)
        self.loose_enforcement = self.get_boolean_parameter(
            'LooseEnforcement', False)
        self.format = self.get_parameter('Format', 'XML')

    @property
    def declared_parameters(self):
        return {"LooseEnforcement": "Whether or not the parsing step should "
                                    "simply store the events (True) or if it "
                                    "should validate the EPCIS events using "
                                    "business rule processing. (False). "
                                    "Default is False",
                "Format": "JSON or XML.  If set to XML, the LooseEnforcement "
                          "parameter is examined.  If set to JSON, inbound "
                          "data must be in the EPCPyYes JSON format and will "
                          "be subject to strong business rule enforcement."
                }

    def execute(self, data, rule_context: RuleContext):
        self.info('Inbound format is configured for %s.', self.format)
        if self.format.upper() == 'XML':
            parser_type = QuartetParser if self.loose_enforcement else BusinessEPCISParser
        else:
            parser_type = JSONParser
        self.info('Loose Enforcement of busines rules set to %s',
                  self.loose_enforcement)
        self.info('Parsing message %s.dat', rule_context.task_name)
        try:
            if isinstance(data, File):
                parser = parser_type(data)
            else:
                parser = parser_type(io.BytesIO(data))
        except TypeError:
            try:
                parser = parser_type(io.BytesIO(data.encode()))
            except AttributeError:
                self.error("Could not convert the data into a format that "
                           "could be handled.")
                raise
        # add the message id to the context
        message_id = parser.parse()
        self.info('Adding Message ID %s to the context under '
                  'key MESSAGE_ID.', message_id)
        rule_context.context[
            ContextKeys.EPCIS_MESSAGE_ID_KEY.value
        ] = message_id
        self.info('Parsing complete.')

    def on_failure(self):
        pass





# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.utils.translation import gettext as _
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from quartet_epcis.models import entries, events, headers


class Command(BaseCommand):
    help = _('Clears out the event data.  For testing and debug systems only.')
    def add_arguments(self, parser):
        parser.add_argument('--force',
                            dest='force',
                            help='Force the removal of the event data.'
                            )
    def handle(self, *args, **options):
        debug = getattr(settings, 'DEBUG', False)
        force = options.get('force', False)
        if debug or force in ['True', 'true', 'TRUE']:
            print('Clearing out event data.')
            entries.Entry.objects.all().delete()
            events.Event.objects.all().delete()
            entries.EntryEvent.objects.all().delete()
            headers.Message.objects.all().delete()
            headers.SBDH.objects.all().delete()
            headers.Partner.objects.all().delete()
            headers.DocumentIdentification.objects.all().delete()
            print('Done.')
        else:
            print('Not valid on non-debug systems.')



from django.contrib import admin

from quartet_epcis.models import entries, events, headers


@admin.register(entries.Entry)
class EntryAdmin(admin.ModelAdmin):
    list_display = (
        'identifier',
        'last_disposition',
        'parent_id',
        'top_id',
        'last_event',
        'last_event_time'
    )
    raw_id_fields = (
        'top_id',
        'parent_id',
        'last_event',
        'last_aggregation_event',
    )
    readonly_fields = (
        'last_event_time',
        'last_aggregation_event_time',
        'is_parent'
    )


@admin.register(entries.EntryEvent)
class EntryEventAdmin(admin.ModelAdmin):
    list_display = (
        'identifier',
        'event',
        'event_type'
    )
    readonly_fields = (
        'event_type',
        'entry',
        'event',
        'is_parent',
        'output',
        'event_time'
    )


@admin.register(events.TransformationID)
class TransformationIDAdmin(admin.ModelAdmin):
    raw_id_fields = ('event',)


@admin.register(events.ErrorDeclaration)
class ErrorDeclarationAdmin(admin.ModelAdmin):
    pass


@admin.register(events.QuantityElement)
class QuantityElementAdmin(admin.ModelAdmin):
    raw_id_fields = ('event',)


@admin.register(events.BusinessTransaction)
class BusinessTransactionAdmin(admin.ModelAdmin):
    readonly_fields = ('event',)


class BusinessTransactionInline(admin.TabularInline):
    model = events.BusinessTransaction
    readonly_fields = ('event',)
    extra = 0


@admin.register(events.InstanceLotMasterData)
class InstanceLotMasterDataAdmin(admin.ModelAdmin):
    raw_id_fields = ('event',)


@admin.register(events.Source)
class SourceAdmin(admin.ModelAdmin):
    pass


@admin.register(events.SourceEvent)
class SourceEventAdmin(admin.ModelAdmin):
    pass


@admin.register(events.Destination)
class DestinationAdmin(admin.ModelAdmin):
    pass


@admin.register(events.DestinationEvent)
class DestinationEventAdmin(admin.ModelAdmin):
    pass

@admin.register(headers.SBDH)
class SBDHAdmin(admin.ModelAdmin):
    raw_id_fields = (
        'document_identification',
    )
    readonly_fields = ('message',)


@admin.register(headers.Partner)
class PartnerAdmin(admin.ModelAdmin):
    raw_id_fields = (
        'header',
    )


@admin.register(headers.DocumentIdentification)
class DocumentIdentificationAdmin(admin.ModelAdmin):
    pass

class ILMDInline(admin.TabularInline):
    model = events.InstanceLotMasterData
    raw_id_fields = ('event',)
    extra = 0

@admin.register(events.Event)
class EventAdmin(admin.ModelAdmin):
    list_display = (
        'action', 'biz_step', 'disposition', 'read_point', 'biz_location',
        'type'
    )
    inlines = [
        BusinessTransactionInline,
        ILMDInline
    ]


def register_to_site(admin_site):
    admin_site.register(entries.Entry, EntryAdmin)
    admin_site.register(entries.EntryEvent, EntryEventAdmin)
    admin_site.register(events.ErrorDeclaration, ErrorDeclarationAdmin)
    admin_site.register(events.Event, EventAdmin)
    admin_site.register(events.Destination, DestinationAdmin)
    admin_site.register(events.Source, SourceAdmin)
    admin_site.register(events.TransformationID, TransformationIDAdmin)
    admin_site.register(events.InstanceLotMasterData,
                        InstanceLotMasterDataAdmin)
    admin_site.register(events.BusinessTransaction, BusinessTransactionAdmin)
    admin_site.register(events.QuantityElement, QuantityElementAdmin)
    admin_site.register(headers.DocumentIdentification, DocumentIdentificationAdmin)
    admin_site.register(headers.Partner, PartnerAdmin)
    admin_site.register(headers.SBDH, SBDHAdmin)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019 SerialLab Corp.  All rights reserved.
import json

from EPCPyYes.core.v1_2 import json_decoders, events as yes_events
from quartet_epcis.models import headers, events
from quartet_epcis.parsing.business_parser import BusinessEPCISParser


class JSONParser(BusinessEPCISParser):

    def parse(self):
        self._message = headers.Message()
        self._message.save()
        if self.stream.startswith('/'):
            with open(self.stream, 'r') as f:
                self.stream = f.read()
        jsonobj = json.loads(self.stream)
        events = jsonobj.get('events', [])
        if len(events) == 0:
            raise self.NoEventsError('There were no events in the inbound'
                                     ' JSON file.')
        for event in events:
            if 'objectEvent' in event:
                decoder = json_decoders.ObjectEventDecoder(event)
                self.handle_object_event(decoder.get_event())
            elif 'aggregationEvent' in event:
                decoder = json_decoders.AggregationEventDecoder(event)
                self.handle_aggregation_event(decoder.get_event())
            elif 'transactionEvent' in event:
                decoder = json_decoders.TransactionEventDecoder(event)
                self.handle_transaction_event(decoder.get_event())
            else:
                raise self.InvalidEventError('The JSON parser encountered an'
                                             ' event that could not be parsed'
                                             ' %s' % str(event))
        self.clear_cache()
        return self._message.id

    class NoEventsError(Exception):
        pass

    class InvalidEventError(Exception):
        pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import logging
from datetime import datetime
from dateutil.parser import parse as parse_date
from django.utils.translation import gettext as _
from typing import List
from django.db.models import QuerySet
from quartet_epcis.db_api.queries import EPCISDBProxy
from quartet_epcis.parsing import errors
from quartet_epcis.parsing.parser import QuartetParser
from quartet_epcis.models import entries, choices, events as db_events
from EPCPyYes.core.v1_2 import events, events as yes_events

logger = logging.getLogger(__name__)

db_proxy = EPCISDBProxy()

EntryList = List[entries.Entry]


class BusinessEPCISParser(QuartetParser):

    def __init__(self, stream, event_cache_size: int = 1024,
                 recursive_decommission: bool = True):
        '''
        Initializes a BusinessEPCISParser.  This parser will enforce business
        rules around aggregation, decommissioning and the like.
        :param stream: The XML message to parse.
        :param event_cache_size: How large the event cache should grow to
        before being committed to the database.
        :param recursive_decommission: Whether or not Entries can be
        implicitly decommissioned when their parent or to Entry is.
        '''
        super().__init__(stream, event_cache_size)
        self.decommissioned_entry_cache = {}
        self.recursive_decommission = recursive_decommission

    def handle_aggregation_event(
        self,
        epcis_event: events.AggregationEvent
    ):
        '''
        Executed when an AggregationEvent xml structure has finished parsing.
        The rules by which AggregationEvents are handled in this module are
        spelled out in section 7.4.3 of the EPCIS 1.2 specification.

        :param epcis_event: An EPCPyYes AggregationEvent instance.
        '''
        logger.debug('Handling an aggregation event.')
        # see what kind of agg event we have here and process accordingly
        if epcis_event.action == events.Action.observe.value:
            # get a list of epcs
            epcs = epcis_event.child_epcs.copy()
            # append the parent to that list if it is there
            if epcis_event.parent_id: epcs.append(epcis_event.parent_id)
            # create entries for each epc if they already do not exist
            self._get_entries(epcs)
            db_event = super().handle_aggregation_event(epcis_event)
        else:
            db_event = self.get_db_event(epcis_event)
            db_event.type = choices.EventTypeChoicesEnum.AGGREGATION.value
            if epcis_event.action == events.Action.add.value:
                self._handle_aggregation_parent(db_event, epcis_event)
            else:
                self._handle_aggregation_delete_action(db_event, epcis_event)
            self._handle_aggregation_entries(db_event, epcis_event)
            self.handle_common_elements(db_event, epcis_event)
            self._append_event_to_cache(db_event)
        return db_event

    def handle_transaction_event(self,
                                 epcis_event: yes_events.TransactionEvent):
        '''
        Checks the epcs for validity before handing off to the base class.
        Validity checks are according to section 7.4.4 of the EPCIS 1.2
        standard.
        :param epcis_event: The EPCPyYes event being parsed.
        '''
        epcs = epcis_event.epc_list.copy()
        if epcis_event.parent_id: epcs.append(epcis_event.parent_id)
        self._get_entries(epcs)
        # if everything is good, hand-off the base class
        db_event = super().handle_transaction_event(epcis_event)

    def handle_object_event(self, epcis_event: yes_events.ObjectEvent):
        '''
        Checks inbound EPCPyYes ObjectEvents and makes sure they follow
        the object event guidelines in section 7.4.2 of the EPCIS 1.2
        standard.
        :param epcis_event: The EPCPyYes object event being scrutinized.
        :return: None
        '''
        if epcis_event.action == events.Action.add.value:
            db_event = super().handle_object_event(epcis_event)
            db_entries = self._get_entries(epcis_event.epc_list)
            self._update_event_entries(db_entries, db_event, epcis_event)
        else:
            db_event = self.get_db_event(epcis_event)
            db_event.type = choices.EventTypeChoicesEnum.OBJECT.value
            db_entries = self._get_entries(epcis_event.epc_list)
            self._update_event_entries(db_entries, db_event, epcis_event)
            if epcis_event.action == events.Action.delete.value:
                self._decommission_entries(db_entries, db_event, epcis_event)
            for db_entry in db_entries:
                entryevent = entries.EntryEvent(
                    entry=db_entry,
                    event_time=epcis_event.event_time,
                    event_type=db_event.type,
                    event=db_event,
                    identifier=db_entry.identifier,
                    output=False
                )
                self.entry_event_cache.append(entryevent)
            self._append_event_to_cache(db_event)
            self.handle_common_elements(db_event, epcis_event)

    def _handle_aggregation_entries(
        self, db_event: str, epcis_event: events.AggregationEvent
    ):
        '''
        Gets the EPCs from the event and caches them for storage in the
        back-end.  Then creates the EntryEvent intersection entity records
        and appends them for storage as well.
        :param db_event: The unique id of the event
        :param epc_list: A list of epcs to be cached.
        :return:
        '''
        if epcis_event.action == events.Action.add.value:
            # check the action date against the latest entry_event
            # if the event is after the last event then update the parent
            # data on the entry
            event_time = epcis_event.event_time
            # get the entries and make sure none are decomissioned or
            # already packed inside another parent
            # step one- check the cache first!
            db_entries, count = self._get_entries_for_aggregation(
                epcis_event)
            # the count should be the same in the event and queryset
            if not count == len(epcis_event.child_epcs):
                error_list = self._get_error_list(epcis_event, db_entries)
                list_msg = ', '.join(error_list)
                raise errors.InvalidAggregationEventError(
                    _('The aggregation event with parent %s '
                      'contained children '
                      'in the child_epc list that were either '
                      'never commissioned, decommissioned '
                      'or already in another aggregation. '
                      'EPC values in question: %s'),
                    epcis_event.parent_id, list_msg
                )
            else:
                # ok, the count matches and they can be packed. now we
                # just check the parent to make sure that the parent is
                # a valid epc
                parent = self._get_entry(epc=epcis_event.parent_id)
                self.create_entry_events(db_entries, db_event, epcis_event)
                self._update_aggregation_entries(db_entries, parent, db_event,
                                                 epcis_event)
        else:
            self.handle_entries(db_event, epcis_event.child_epcs,
                                epcis_event)

    def _get_error_list(self, epcis_event: events.AggregationEvent,
                        db_entries: EntryList):
        """
        If there is a commissioning issue trying to pack items, we will
        probably want to know which EPC is the problem.  This will compare
        the list from the cache and compare it to the epc list in an
        aggregation event and spit out the epcs that were not in the cache
        and / or database.
        :param epcis_event: The aggregation event in question.
        :param db_entries: The entries returned from the cache.
        :return: A list of EPCs that were not in the cache or database.
        """
        entry_list = []
        for entry in db_entries:
            entry_list.append(entry.identifier)
        epc_list = epcis_event.child_epcs
        return list(set(epc_list) - set(entry_list))

    def create_entry_events(self, db_entries, db_event, epcis_event):
        '''
        For a list of Entry instance and an Event instance, will create
        the intersection entity model instances and add them to the
        entry event cache.
        :param db_entries: A list of Entry model instances.
        :param db_event: A events.Event model instance.
        :param epcis_event: An EPCPyYes event.
        :return: None
        '''
        for db_entry in db_entries:
            entry_event = entries.EntryEvent(
                entry=db_entry,
                event_time=epcis_event.event_time,
                event=db_event,
                event_type=db_event.type,
                identifier=db_entry.identifier
            )
            self.entry_event_cache.append(entry_event)

    def _get_entries_for_aggregation(self,
                                     epcis_event: events.AggregationEvent):
        '''
        Will pull entries first from the local entries cache then, if not
        found, will attempt to get them from the database.
        :param epcis_event: The event with the epcs to look for in the Entries
        cache and database model.
        :return: A two-tuple with a list or queryset of Entry model instances
        and an integer expressing the number of entries found.
        '''
        db_entries = []
        count = 0
        parent = self._get_entry(epcis_event.parent_id)
        # try the local cache
        for epc in epcis_event.child_epcs:
            entry = self.entry_cache.get(epc)
            if entry and not entry.decommissioned and not entry.parent_id:
                db_entries.append(entry)
        count = len(db_entries)
        # if nothing was found, try the database
        if count < len(epcis_event.child_epcs):
            kwargs = {'identifier__in': epcis_event.child_epcs,
                      'decommissioned': False,
                      'parent_id': None}
            db_entries = entries.Entry.objects.select_for_update().filter(
                **kwargs
            )
            count = db_entries.count()
        return db_entries, count

    def _update_aggregation_entries(
        self,
        db_entries,
        parent: entries.Entry,
        db_event: db_events.Event,
        epcis_event: events.AggregationEvent
    ):
        '''
        Will update the value of the entries and add them to the cache if they
        were not there already or just update the cache entries.  If a list
        is sent in then we assume the entries are already in the cache.
        If a queryset is sent in then we assume the entries are not in the
        cache and attempt to add them.
        :param entries: A queryset or list of entries.
        :return: None
        '''
        # the top is the parent's top or the parent itself...
        for db_entry in db_entries:
            db_entry.last_aggregation_event_time = epcis_event.event_time
            db_entry.last_aggregation_event = db_event
            db_entry.last_aggregation_event_action = epcis_event.action
            if parent:
                db_entry.top_id = parent.top_id or parent
            else:
                db_entry.top_id = parent
            db_entry.last_event = db_event
            db_entry.last_event_time = self._parse_date(epcis_event)
            db_entry.last_disposition = epcis_event.disposition
            db_entry.parent_id = parent
            # db_entry.save()
            if db_entry.is_parent:
                # self._update_aggregation_entries(
                #     entries.Entry.objects.filter(parent_id=db_entry),
                #     db_entry, db_event, epcis_event)
                self._update_aggregation_entries(
                    self._get_child_entries(db_entry),
                    db_entry, db_event, epcis_event)
            # make sure to keep in cache
            self.entry_cache[db_entry.identifier] = db_entry

    def _parse_date(self, epcis_event):
        event_time = parse_date(epcis_event.event_time)
        return event_time

    def _get_child_entries(self, db_entry: entries.Entry):
        '''
        Gets child entries for an entry from the cache first then
        from the database in not found.
        :param db_entry: An entry marked with is_parent = True
        :return: A list or QuerySet of child entries.
        '''
        # look in the cache first
        ret = []
        for k, v in self.entry_cache.items():
            if v.parent_id == db_entry:
                ret.append(v)
        # now get any from the db
        db_children = db_proxy.get_entries_by_parent(db_entry)
        for db_child in db_children:
            ret.append(db_child)
        return ret

    def _update_event_entries(
        self,
        db_entries,
        db_event: db_events.Event,
        epcis_event: events.EPCISEvent
    ):
        '''
        Will update the value of the entries and add them to the cache if they
        were not there already or just update the cache entries.  If a list
        is sent in then we assume the entries are already in the cache.
        If a queryset is sent in then we assume the entries are not in the
        cache and attempt to add them.
        :param entries: A queryset or list of entries.
        :return: None
        '''
        # the top is the parent's top or the parent itself...
        if isinstance(db_entries, list):
            for db_entry in db_entries:
                db_entry.last_event = db_event
                db_entry.last_event_time = self._parse_date(epcis_event)
                db_entry.last_disposition = epcis_event.disposition
                # db_entry.save()
        elif isinstance(db_entries, QuerySet):
            # update the database
            count = db_entries.update(
                last_event=db_event,
                last_event_time=self._parse_date(epcis_event),
                last_disposition=epcis_event.disposition
            )
            if count == 0:
                raise errors.EntryException(
                    _('No Entry records were updated.')
                )
            # then update the local cache if the db update was successful
            for entry in db_entries:
                self.entry_cache[entry.identifier] = entry

    def _handle_aggregation_parent(self, db_event: db_events.Event,
                                   epcis_event: events.AggregationEvent):
        '''
        Will retrieve or create and store the parent and or/top level
        id as an Entry in the
        entry cache, create an EntryEvent and store that in the EntryEvent
        cache and return the new or existing Entry.

        :param top_id: The parent epc of an aggregation event.
        :param db_event: The Event model instance.
        :param epcis_event: The EPCPyYes event with the inbound event data.
        :return: An Entry model instance updated to reflect the current
        event being processed.
        '''
        # check the cache or get from the db
        try:
            entry = self._get_entry(epcis_event.parent_id)
        except errors.EntryException:
            raise errors.EntryException(
                _('The parent entry with identifer %s was either '
                  'decommissioned or was never commissioned.'),
                epcis_event.parent_id
            )
        # set all the pointers and convienince properties on the entry
        entry.last_aggregation_event_action = epcis_event.action
        entry.last_aggregation_event_time = self._parse_date(epcis_event)
        entry.last_aggregation_event = db_event
        entry.is_parent = True
        entry.last_event = db_event
        entry.last_event_time = self._parse_date(epcis_event)
        entry.last_disposition = epcis_event.disposition
        # entry.save()
        # if its not in the cache it needs to be added
        self.entry_cache[entry.identifier] == entry
        # create an entry event and add to the cache
        self._create_parent_entry_event(db_event, epcis_event)
        logger.debug('Cached Entry for top id %s', epcis_event.parent_id)
        return entry

    def _handle_aggregation_delete_action(self, db_event: db_events.Event,
                                          epcis_event: events.AggregationEvent):
        # 1. see if there is a parent and children
        # if parent and children then set the parent and top of the children
        # to none
        if epcis_event.parent_id and len(epcis_event.child_epcs) > 0:
            # ok, we have parent and children
            db_entries = self._get_entries(
                epcis_event.child_epcs
            )
            # create the entry events for the children
            self.create_entry_events(db_entries, db_event, epcis_event)
        else:
            db_entries = db_proxy.get_entries_by_parent(epcis_event.parent_id)
            # clear out any entries that have these as top_id
            lower_entries = db_proxy.get_entries_by_top(
                self._get_entry(epcis_event.parent_id)
            )
            lower_entries.update(
                top_id=None
            )
        self._create_parent_entry_event(db_event, epcis_event)

        self._update_aggregation_entries(
            db_entries, None, db_event, epcis_event
        )

    def _create_parent_entry_event(self, db_event, epcis_event):
        '''
        Based on the EPCPyYes event and the database Event model instance,
        will create an intersection entity reference to the parent of the
        event and the parent entry.
        :param db_event: The Event model instance.
        :param epcis_event: An EPCPyYes event with a parent or top.
        :return: Returns the created EntryEvent model instance.
        '''
        # create the entry event for the parent
        entryevent = entries.EntryEvent(
            entry=self._get_entry(epcis_event.parent_id),
            event=db_event,
            event_time=epcis_event.event_time,
            event_type=db_event.type,
            identifier=epcis_event.parent_id,
            is_parent=True
        )
        self.entry_event_cache.append(entryevent)
        return entryevent

    def _get_entry(self, epc: str):
        '''
        Will look for an entry in the cache and then, if not found, from the
        database.  If it gets the entry from the database, it will then add
        it to the local cache.
        :param epc: The entry identifier to search for.
        :return: An Entry model instance.
        '''
        entry = self.entry_cache.get(epc)
        if entry and entry.decommissioned:
            raise errors.DecommissionedEntryException(
                'The entry with identifier %s has been decommissioned.',
                epc
            )
        if not entry:
            try:
                entry = entries.Entry.objects.get(
                    identifier=epc,
                    decommissioned=False
                )
                # add it to the cache
                self.entry_cache[epc] = entry
            except entries.Entry.DoesNotExist:
                raise errors.EntryException(
                    _('The entry with identifier %s could '
                      'not be found.  It was either '
                      'decommissioned or never commissioned'),
                    epc
                )
        return entry

    def _get_entries(self, epcs: list):
        '''
        Pulls first from the cache and then from the database.
        :param epcs: The epcs to use for lookup of entries.
        :return: A list or queryset of entries.
        '''
        db_entries = []
        for epc in epcs:
            epc = self.entry_cache.get(epc)
            if epc: db_entries.append(epc)
        if len(db_entries) != len(epcs):
            db_entries = db_proxy.get_entries_by_epcs(
                epcs,
                select_for_update=False
            )
            if db_entries.count() != len(epcs):
                raise errors.EntryException(
                    _('Invalid Entry.  One of the values in the '
                      'event has either '
                      'been decommissioned or was '
                      'never commissioned.')
                )
        return db_entries

    def _decommission_entries(
        self,
        db_entries: EntryList,
        db_event: db_events.Event,
        epcis_event: events.EPCISEvent,
        recursive: bool = True
    ):
        '''
        Will mark each entry in the entry list as decommissioned.
        :param db_entries: The entries to set as decommissioned.
        :param recursive: Whether or not to decommission any child entries.
        Default = True
        '''
        if recursive:
            children = db_proxy.get_entries_by_parents(db_entries)
            if children.count() > 0:
                self._decommission_entries(children, db_event, epcis_event)
        for entry in db_entries:
            entry.decommissioned = True
            entry.last_event = db_event
            entry.last_event_time = self._parse_date(epcis_event)
            entry.last_disposition = epcis_event.disposition
            # remove from the main cache (so it can't be selected later)
            # and add to the decommissioned entry cache
            self.entry_cache.pop(entry.identifier, None)
            self.decommissioned_entry_cache[entry.identifier] = entry
            entry_event = entries.EntryEvent(entry=entry,
                                             event_time=epcis_event.event_time,
                                             event_type=db_event.type,
                                             event=db_event,
                                             identifier=entry.identifier,
                                             output=False)
            self.entry_event_cache.append(entry_event)

    def clear_cache(self):
        # create events
        event_cache = self._get_sorted_event_cache()
        db_events.Event.objects.bulk_create(event_cache)
        # update entries
        for db_entry in list(self.entry_cache.values()):
            db_entry.save()
        # clear the event cache
        self.event_cache.clear()
        decommissioned_entries = list(
            self.decommissioned_entry_cache.values())
        for decommissioned_entry in decommissioned_entries:
            decommissioned_entry.save()
        decommissioned_entries.clear()
        super().clear_cache()





# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
"""

A series of Django two-tuple choices for models and also contains
an EventType enum for EPCIS event types to ensure that the code and the
choices always match up accurately.

"""


from enum import Enum

ACTION_CHOICES = (
    ('ADD', 'Add'),
    ('OBSERVE', 'Observe'),
    ('DELETE', 'Delete')
)

EVENT_TYPE_CHOICES = (
    ('ag', 'Aggregation'),
    ('ob', 'Object'),
    ('tx', 'Transaction'),
    ('tf', 'Transformation')
)

class EventTypeChoicesEnum(Enum):
    '''
    Helper used to avoid any errors when using the choices outside of models.
    '''
    AGGREGATION = 'ag'
    OBJECT = 'ob'
    TRANSACTION = 'tx'
    TRANSFORMATION = 'tf'
